""" loopy instructions to mark stages of computations """

from dune.perftool.generation import (generator_factory,
                                      noop_instruction,
                                      )

from loopy import add_dependency
from loopy.match import Id


@generator_factory(item_tags=("stage",), cache_key_generator=lambda n, **kw: n)
def stage_insn(n, **kwargs):
    assert 'id' not in kwargs

    # Get an ID for this instruction
    id = 'stage_insn_{}'.format(n)

    # Chain dependencies of stage instructions
    if n > 0:
        kwargs['depends_on'] = kwargs.get('depends_on', frozenset([])).union(frozenset([stage_insn(n - 1, **kwargs)]))

    # Actually issue the instruction
    noop_instruction(id=id, **kwargs)

    return id


def finalize_stage_instructions(kernel):
    for i in range(len(stage_insn._memoize_cache)):
        deps = frozenset({insn.id for insn in kernel.instructions if stage_insn(i) in insn.depends_on and not insn.id.startswith('stage_insn_')})
        for dep_id in deps:
            kernel = add_dependency(kernel, Id(stage_insn(i + 1)), dep_id)

    return kernel
