from dune.perftool.loopy.temporary import DuneTemporaryVariable
from dune.perftool.generation.loopy import AccumulationGlobalArg

from loopy.target import (TargetBase,
                          ASTBuilderBase,
                          DummyHostASTBuilder,
                          )
from loopy.target.c import CASTBuilder
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper, CExpressionToCodeMapper
from loopy.symbolic import FunctionIdentifier

from pymbolic.primitives import Call, Subscript, Variable


_registry = {'float32': 'float',
             'int32': 'int',
             'float64': 'double',
             'string': 'std::string',
             'str': 'std::string'}


class MyMapper(ExpressionToCExpressionMapper):
    def map_subscript(self, expr, type_context):
        arr = self.find_array(expr)
        if isinstance(arr, DuneTemporaryVariable) and not arr.managed:
            # If there is but one index, we do not need to handle this
            if isinstance(expr.index, (Variable, int)):
                return expr

            # Else, we construct a nested Subscript chain
            ret = expr.aggregate
            for i in expr.index:
                ret = Subscript(ret, i)
            return ret
        elif isinstance(arr, AccumulationGlobalArg):
            assert isinstance(arr.transform, FunctionIdentifier)
            pseudo_subscript = Subscript(expr.aggregate, expr.index)
            flattened = ExpressionToCExpressionMapper.map_subscript(self, pseudo_subscript, type_context)
            transformed = ExpressionToCExpressionMapper.map_call(self, Call(arr.transform, (flattened.index,)), type_context)
            return Subscript(flattened.aggregate + '.base()', transformed)
        else:
            return ExpressionToCExpressionMapper.map_subscript(self, expr, type_context)


class DuneASTBuilder(CASTBuilder):
    def get_expression_to_c_expression_mapper(self, codegen_state):
        return MyMapper(codegen_state)

    def get_temporary_decl(self, knl, schedule_index, temp_var, decl_info):
        # If this is not a DuneTemporaryVariable, it was introduced by loopy
        # and it should be totally under loopys control: Call the base class implementation!
        if not (isinstance(temp_var, DuneTemporaryVariable) and temp_var.custom_declaration):
            return CASTBuilder.get_temporary_decl(self, knl, schedule_index, temp_var, decl_info)

        from cgen import Line
        if temp_var.decl_method:
            return Line(temp_var.decl_method(temp_var.name, temp_var.shape, temp_var.shape_impl))


class DuneTarget(TargetBase):
    def __init__(self):
        # Set fortran_abi to allow reusing CASTBuilder for the moment
        self.fortran_abi = False

    def split_kernel_at_global_barriers(self):
        return False

    def get_host_ast_builder(self):
        return DummyHostASTBuilder(self)

    def get_device_ast_builder(self):
        return DuneASTBuilder(self)

    def dtype_to_typename(self, dtype):
        # For now, we do this the simplest possible way
        return _registry[dtype.dtype.name]

    def is_vector_dtype(self, dtype):
        return False
